// Package `tracking` defines a generic track (sort of record) type
// and a container type that allows storing of tracks in a SQL database.
package tracking

import (
	sqllib "database/sql"
	"encoding/json"
	"fmt"
	"time"

	lib "git.sr.ht/~cco/go-scopes"
	"git.sr.ht/~cco/go-scopes/logging/log"
	"git.sr.ht/~cco/go-scopes/storage"
	sql "git.sr.ht/~cco/go-scopes/storage"
)

type ItemFactory func(*Container, ...string) *Track

// basic track implementation

type Track struct {
	trackId   lib.Ident
	Head      lib.StrMap
	TimeStamp *time.Time
	Data      lib.Map
	container *Container
}

func (tr *Track) TrackId() lib.Ident {
	return tr.trackId
}

func (tr *Track) Container() *Container {
	return tr.container
}

func MakeTrack(cont *Container, h ...string) *Track {
	tr := Track{
		Head:      lib.StrMap{},
		container: cont,
	}
	tr.SetHead(h...)
	return &tr
}

func (tr *Track) SetHead(h ...string) {
	for i, k := range tr.container.HeadFields {
		if i >= len(h) {
			break
		}
		if h[i] != "" {
			tr.Head[k] = h[i]
		}
	}
}

func (tr *Track) ScanP(rows *sql.Rows) error {
	tr.Head = lib.StrMap{}
	var d []any
	for _, _ = range tr.container.HeadFields {
		//tr.Head[k] = ""
		var hv string
		d = append(d, &hv)
	}
	var ts string
	d = append(d, &ts)
	var rd string
	d = append(d, &rd)
	err := rows.Scan(d...)
	fmt.Printf("%+v\n", d)
	fmt.Printf("%+v, %+v, %+v, %+v\n", *d[0].(*string), *d[1].(*string), *d[2].(*string), *d[3].(*string))
	return err
}

// basic container implementation

type ContDef struct {
	ItemFactory ItemFactory
	TableName   string
	HeadFields  []string
	Indexes     [][]string
}

type Container struct {
	*ContDef
	Storage *sql.Storage
}

func Tracks(db *sql.Storage) *Container {
	return &Container{container_definition, db}
}

func (cont *Container) Get(id lib.Ident) *Track {
	db := cont.Storage
	var trs []*Track
	processTracks := func(r *sql.Rows) error {
		tr := cont.ItemFactory(cont)
		err := tr.ScanP(r)
		trs = append(trs, tr)
		return err
	}
	qu := lib.Map{
		"schema":    db.Schema,
		"tablename": cont.TableName,
		"scols":     append(cont.HeadFields, "timestamp", "data"),
		"qucols":    lib.StrSlice{"trackid"},
	}
	sql := storage.BuildSql(SqlSelect, qu)
	print(sql)
	//tr := cont.ItemFactory(cont)
	db.Query(processTracks, sql, id)
	if len(trs) > 0 {
		return trs[0]
	}
	return nil
}

func (cont *Container) Query(headValues lib.StrSlice) *Track {
	db := cont.Storage
	data := lib.Map{
		"schema":    db.Schema,
		"tablename": cont.TableName,
		"scols":     lib.StrSlice{"taskid", "data"},
		"qucols":    lib.StrSlice{"taskid", "username"},
		"ordcols":   lib.StrSlice{"timestamp"},
	}
	sql := storage.BuildSql(SqlSelect, data)
	//print(sql)
	_ = sql
	tr := cont.ItemFactory(cont)
	return tr
}

func (cont *Container) NewTrack(h []string, data lib.Map) *Track {
	tr := cont.ItemFactory(cont, h...)
	//tr.SetData(data)
	cont.Insert(tr)
	return tr
}

func (cont *Container) Save(t *Track) lib.Ident {
	return 0
}

func (cont *Container) Insert(tr *Track) lib.Ident {
	var columns []string
	var values []any
	for _, k := range cont.HeadFields {
		columns = append(columns, k)
		values = append(values, tr.Head[k])
	}
	columns = append(columns, "Data")
	b, _ := json.Marshal(tr.Data)
	values = append(values, b)
	db := cont.Storage
	data := lib.Map{
		"schema":    db.Schema,
		"tablename": cont.TableName,
		"columns":   columns,
	}
	sql := storage.BuildSql(SqlInsert, data)
	var trid lib.Ident
	var ts *time.Time
	var tsstr string
	proc := func(r *sqllib.Rows) error {
		err := r.Scan(&trid, &ts)
		if err != nil {
			err = r.Scan(&trid, &tsstr)
			ts = ParseDateTime(tsstr)
		}
		return err
	}
	if err := db.Query(proc, sql, values...); err == nil {
		tr.trackId = trid
		tr.TimeStamp = ts
		return trid
	}
	return 0
}

func ParseDateTime(inp string) *time.Time {
	ts, err := time.Parse("2006-01-02 15:04:05", inp)
	if err == nil {
		//fmt.Printf("%+v\n", ts)
		return &ts
	}
	log.Error(err).Msg("storage.tracking.ParseDateTime")
	return nil
}

func (cont *Container) CreateTable() {
	db := cont.Storage
	data := lib.Map{
		"schema":     db.Schema,
		"tablename":  cont.TableName,
		"headFields": cont.HeadFields,
		"indexes":    cont.Indexes,
		"params":     db.Params,
	}
	sql := storage.BuildSql(SqlCreate, data)
	if _, err := db.Exec(sql); err != nil {
		panic(err)
	}
}

// container definition

var container_definition *ContDef

func init() {
	container_definition = &ContDef{
		ItemFactory: MakeTrack,
		TableName:   "tracks",
		HeadFields:  []string{"taskId", "userName"},
		Indexes: [][]string{
			[]string{"taskId", "userName"},
			[]string{"userName"},
		},
	}
}
